////////////////////////////////////////////////////////////////////////
// Class:       NoiseRemoval
// Plugin Type: producer (art v3_05_01)
// File:        NoiseRemoval_module.cc
//
// Generated at Thu Jun 25 12:17:06 2020 by Wanwei Wu using cetskelgen
// from cetlib version v3_10_00.
// 
// Author: Wanwei Wu, wwu@fnal.gov
//
// About: remove coherent noise and exponential tail noise
//
// a) coherent noise:
//      1) correlation between channels are studied;
//      2) two options: 
//           - default 24 channels/group for both plane
//           - channel group based on correlation study (correlation
//             is time-dependent)
//      3) current coherent noise removal is based on correlation study
//         using an empty data event (Run 806, event 19508).
// b) exponential tail removal:
//      1) two types of exponential tails: from electronics response and 
//      cable impedance
//      2) two methods includes:
//           - baseline subtraction (works for both tails)
//           - undershoot tail removal (not perfect, may only model the 
//           electronics response, just keep it here)
//      3) exponential tail removal only happens for collection plane here
// c) noise removal strategy: removal exponential tails first, then do 
//    the coherent noise removal
// d) RawDigit contains all 480 channles. For recob::Wire after NoiseRemoval,
//    only valid channel are saved. Those bad channels will not be saved.
////////////////////////////////////////////////////////////////////////

// art framework includes
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "canvas/Utilities/InputTag.h"
#include "canvas/Persistency/Common/Ptr.h"
#include "canvas/Persistency/Common/Assns.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "cetlib_except/exception.h"
#include "cetlib/search_path.h"

#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art_root_io/TFileService.h"
#include "art_root_io/TFileDirectory.h"

// larsoft includes
#include "lardataobj/RawData/raw.h"
#include "lardataobj/RawData/RawDigit.h"
#include "lardataobj/RecoBase/Wire.h"
#include "lardata/ArtDataHelper/WireCreator.h"
#include "lardata/Utilities/LArFFT.h"
#include "lardata/Utilities/AssociationUtil.h"
#include "lardata/Utilities/SignalShaping.h"
#include "lardata/DetectorInfoServices/DetectorPropertiesService.h"
#include "larcore/Geometry/Geometry.h"
#include "larcoreobj/SimpleTypesAndConstants/RawTypes.h" // raw::ChannelID_t

#include "larevt/CalibrationDBI/Interface/ChannelStatusService.h"
#include "larevt/CalibrationDBI/Interface/ChannelStatusProvider.h"

// ArgoNeuT includes
#include "T962Utilities/SignalShapingServiceT962.h"

#include <memory>

// ROOT includes
#include "TFile.h"
#include "TH1D.h"
#include "TMath.h"

using namespace std;


class NoiseRemoval;


class NoiseRemoval : public art::EDProducer {
public:
  explicit NoiseRemoval(fhicl::ParameterSet const& p);
  // The compiler-generated destructor is fine for non-base
  // classes without bare pointers or other resource use.

  // Plugins should not be copied or assigned.
  NoiseRemoval(NoiseRemoval const&) = delete;
  NoiseRemoval(NoiseRemoval&&) = delete;
  NoiseRemoval& operator=(NoiseRemoval const&) = delete;
  NoiseRemoval& operator=(NoiseRemoval&&) = delete;

  // Required functions.
  void produce(art::Event& e) override;
  void beginJob() override;

private:

  // function: coherent noise removal
  std::vector<float> getMedianCorrection(std::vector< std::vector<float> >& holder, unsigned int dataSize, unsigned short plane);
  std::vector<bool> roiMask(std::vector<float>& channel, unsigned short plane);

  // function: exponential tails removal
  void estimatepars(std::vector<float> &holder, double &pedoffset, double &tail0, int channel);
  void setData(std::vector<float> & holder, std::vector<float> & signal, std::vector<float> & tail, double tail0, double ped, int channel);
  void SubtractBaseline(std::vector<float>& holder, int baseSampleBins);

  // configuration from fhicl file
  std::string fDigitModuleLabel;
  std::string fRawDigitInstance;
  std::vector<double> fRoiThresholdLow; // threshold to check sig/noise low
  std::vector<double> fRoiThresholdHigh; // threshold to check sig/noise high
  int fRoiPadLow; // signal-like region
  int fRoiPadHigh; // signal-like region
  bool fRemoveCoherentNoise;
  bool fCoherentNoiseDefault; // true: default coherent noise removal; false: correlation based coherent noise removal 
  std::vector<int> fNumChannelGroupPlane0;
  std::vector<int> fNumChannelGroupPlane1;
  std::vector<bool> fRemoveUndershootNoise; // for each plane
  bool fBaselineSubtractOnly; // true: only baseline subtract; false: do undershoot tail removal first, then baseline subtract
  std::vector<int> fBaseSampleBins;
  std::string fResponseFile;
  bool fDoDeconvolution;
  bool fDodQdxCalib; // true: will apply wire-by-wire calibration
  int fCheckChannel; // get waveform for analysis
  double fADCCut;

  std::map<unsigned int, float> fdQdxCalib; // Map to do wire-by-wire calibration, key is channel number, content is correction factor

  int fNticks;

  TH1D *h1_adc_raw;
  TH1D *h1_adc_raw_tailremoval;
  TH1D *h1_adc_raw_tailremoval_coherentremoval;
  
  TH1D *h1_decayconstant;
};


NoiseRemoval::NoiseRemoval(fhicl::ParameterSet const& p)
  : EDProducer{p}  // ,
  // More initializers here.
{

  cet::search_path sp("FW_SEARCH_PATH");

  fDigitModuleLabel       = p.get<std::string>("DigitModuleLabel");
  fRawDigitInstance       = p.get<std::string>("RawDigitInstance");
  fRoiThresholdLow        = p.get<std::vector<double>>("RoiThresholdLow");
  fRoiThresholdHigh       = p.get<std::vector<double>>("RoiThresholdHigh");
  fRoiPadLow              = p.get<int>("RoiPadLow");
  fRoiPadHigh             = p.get<int>("RoiPadHigh");
  fRemoveCoherentNoise    = p.get<bool>("RemoveCoherentNoise");
  fCoherentNoiseDefault   = p.get<bool>("CoherentNoiseDefault");
  fNumChannelGroupPlane0  = p.get<std::vector<int>>("NumChannelGroupPlane0");
  fNumChannelGroupPlane1  = p.get<std::vector<int>>("NumChannelGroupPlane1");
  fRemoveUndershootNoise  = p.get<std::vector<bool>>("RemoveUndershootNoise");
  fBaselineSubtractOnly   = p.get<bool>("BaselineSubtractOnly");
  fBaseSampleBins         = p.get<std::vector<int>>("BaseSampleBins");
  fResponseFile           = sp.find_file(p.get<std::string>("ResponseFile"));
  fDoDeconvolution        = p.get<bool>("DoDeconvolution");
  fDodQdxCalib            = p.get<bool>("DodQdxCalib", false);
  fCheckChannel           = p.get<int>("CheckChannel");
  fADCCut                 = p.get<double>("ADCCut");

  // Call appropriate produces<>() functions here.
  // Call appropriate consumes<>() for any products to be retrieved by this module.
  produces< std::vector<recob::Wire> >();
  produces<art::Assns<raw::RawDigit, recob::Wire>>();

  if (fRemoveCoherentNoise) std::cout << "NoiseRemoval: will remove coherent noise." << std::endl;
  if (fRemoveUndershootNoise[0]) std::cout << "NoiseRemoval: will remove exponential tail noise on plane 0." << std::endl;
  if (fRemoveUndershootNoise[1]) std::cout << "NoiseRemoval: will remove exponential tail  noise on plane 1." << std::endl;
 
  if (fDodQdxCalib) {
    std::string fileName = p.get< std::string >("dQdxCalibFileName");
    std::string fullname;
    sp.find_file(fileName, fullname);

    if (fullname.empty()) {
      throw cet::exception("NoiseRemoval") << "dQdxCalib File not found!";
    }
    else {
      mf::LogInfo("NoiseRemoval") << "Applying wire-by-wire calibration using file " << fileName;
    }

    std::ifstream inFile(fullname, std::ios::in);
    std::string line;

    while (std::getline(inFile,line)) {
      unsigned int channel;
      float        constant;
      std::stringstream linestream(line);
      linestream >> channel >> constant;
      fdQdxCalib[channel] = constant;
      if (channel%100 == 0) std::cout << "Channel " << channel << " correction factor " << fdQdxCalib[channel] << std::endl;
    }
  }

  // DetectorPropertiesService
  auto const detProp = art::ServiceHandle<detinfo::DetectorPropertiesService const>()->DataForJob();
  fNticks = detProp.ReadOutWindowSize(); // number of clock ticks per event
  //std::cout << "Number of clock ticks per event: " << fNticks << std::endl;
}


void NoiseRemoval::produce(art::Event& e)
{
  // Implementation of required member function here.

  art::ServiceHandle<geo::Geometry const> fGeometry;
  art::ServiceHandle<util::SignalShapingServiceT962> sss;
  art::ServiceHandle<util::LArFFT> fFFT;
  int transformSize = fFFT->FFTSize();
  //cout << "transformSize: " << transformSize << endl;

  // channel status
  lariov::ChannelStatusProvider const& channelStatus
    = art::ServiceHandle<lariov::ChannelStatusService const>()->GetProvider();
  
  // get RawDigit
  art::Handle< std::vector<raw::RawDigit> > rawdigitListHandle;
  std::vector< art::Ptr<raw::RawDigit> > rawdigitlist;
  if (e.getByLabel(fDigitModuleLabel, fRawDigitInstance, rawdigitListHandle)) {
    art::fill_ptr_vector(rawdigitlist, rawdigitListHandle);
  }
  
  // make a collection of Wires and an association set
  std::unique_ptr<std::vector<recob::Wire> > wirecol(new std::vector<recob::Wire>);
  std::unique_ptr<art::Assns<raw::RawDigit,recob::Wire> > WireDigitAssn(new art::Assns<raw::RawDigit,recob::Wire>);

  // ------ raw adc with pedestal subtracted -------
  std::vector<std::vector<float>> rawAdcVec(480); // [480][2048]
  // loop over all channels/wires
  for (unsigned int i=0; i<rawdigitlist.size(); ++i) {
    art::Ptr<raw::RawDigit> rawdigit = rawdigitlist[i];
 
    // number of samples in uncompressed ADC data
    int nSamples =  rawdigit->Samples();
  
    // channel number
    int channel = rawdigit->Channel();

    raw::RawDigit::ADCvector_t rawadc(nSamples);  // vector holding uncompressed adc values
    raw::Uncompress(rawdigit->ADCs(), rawadc, rawdigit->Compression());
    
    // pedestal
    float pedestal = rawdigit->GetPedestal();

    std::vector<float> holder(nSamples, 0);
    for (int j=0; j<nSamples; j++) {
      holder[j] = rawadc[j] - pedestal;
    }
    
    if (channel==fCheckChannel) {
      for (int j=0; j<nSamples; j++) {
        h1_adc_raw->SetBinContent(j+1, holder[j]);
      }
    }

    int plane = 0;
    if (channel >=240) {
      plane = 1;
    }

    // ----remove exponential tails ----------
    if (fRemoveUndershootNoise[plane]) {
      
      if (!fBaselineSubtractOnly) {
        // starts with ADC in holder[2048]; end with newholder[2048]
        std::vector<float> newholder(nSamples);
        double pedoffset = 0;
        double tail0 = 0;

        estimatepars(holder, pedoffset, tail0, channel);

        std::vector<float> vec_tail(nSamples);
        setData(holder, newholder, vec_tail, tail0, pedoffset, channel);

        for (int j=0; j<nSamples; j++) {
          holder[j] = newholder[j];
        }
      }

      SubtractBaseline(holder, fBaseSampleBins[plane]); // for unremoval tail base adjustment

      if (channel==fCheckChannel) {
        for (int j=0; j<nSamples; j++) {
          h1_adc_raw_tailremoval->SetBinContent(j+1, holder[j]);
        }
      }

    } // fRemoveUndershootNoise[plane]


    rawAdcVec[channel] = holder;
  } // loop over rawdigitlist i


  // ---------- coherent noise ------
  std::vector<int> channelgroupvec;
  // default:divide channels into 20 groups, find noise ADC median among channels in group for each tick (there are 2048 ticks in total), the noise ADC median is called coherent noise
  // correlation based method: channels are grouped according the correlation
  if (fCoherentNoiseDefault) {
    int numchgrps = 20; // channel groups
    int numchpergrp = 24; // channels per group
    for (int k=0; k<numchgrps; k++) {
      channelgroupvec.push_back(numchpergrp);
    }
  }
  else {
    for (size_t k = 0; k < fNumChannelGroupPlane0.size(); k++) {
      channelgroupvec.push_back(fNumChannelGroupPlane0[k]);
    }
    for (size_t k = 0; k < fNumChannelGroupPlane1.size(); k++) {
      channelgroupvec.push_back(fNumChannelGroupPlane1[k]);
    }
  }

  unsigned int nchgrps =channelgroupvec.size();// channel groups
  unsigned int nchpergrp; // channels per group
  unsigned int channels_previous = 0; // num channels in previous (n-1) groups

  std::vector< std::vector<float> > cohNoise; // [nchgrps][2048]

  // loop through groups of channels
  for (unsigned int chgrp = 0; chgrp < nchgrps; chgrp++) {
    
    nchpergrp = channelgroupvec[chgrp];
    if (chgrp != 0) {
      channels_previous += channelgroupvec[chgrp-1];
    }
    
    //cout << "chgrp: " << chgrp << "; channelgroupvec: " << channelgroupvec[chgrp] << endl;

    std::vector< std::vector<float> > channelgroup; // [channels][2048]: exclude bad channels in the group

    unsigned short plane = 0;
    unsigned int dataSize = fNticks;
    
    for (unsigned int ch = 0; ch < nchpergrp; ++ch) {
      unsigned int channel = channels_previous + ch;

      if(channelStatus.IsBad(channel)) continue;

      if (channel >= 240) plane = 1;

      channelgroup.push_back(rawAdcVec[channel]);
    } // loop over ch
    
    auto correction = getMedianCorrection(channelgroup, dataSize, plane);
    cohNoise.push_back(correction);

  } // loop over nchgrps chgrp

  // loop over all channels/wires
  for (unsigned int i=0; i<rawdigitlist.size(); ++i) {
    art::Ptr<raw::RawDigit> rawdigit = rawdigitlist[i];
    // channel number
    int channel = rawdigit->Channel();

    int plane = 0;
    if (channel>240) plane = 1;

    if(channelStatus.IsBad(channel)) continue; // do not save bad channels

    // number of samples in uncompressed ADC data
    int nSamples =  rawdigit->Samples();

     std::vector<float> holder = rawAdcVec[channel];

    // remove coherent noise
    if (fRemoveCoherentNoise) {
      int chgroup = -1;
      int channels_1 = 0; // up to group chgrp
      int channels_2 = 0; // up to group chgrp+1
      for (unsigned int chgrp = 0; chgrp < nchgrps; ++chgrp) {
        if (chgrp !=0) {
          channels_1 += channelgroupvec[chgrp-1];
        } 
        channels_2 += channelgroupvec[chgrp];

        if (channel >= channels_1 && channel <channels_2) {
          chgroup = chgrp;
          break;
        }
      }
      
      if (chgroup == -1 || chgroup >=int(nchgrps)) {
        throw cet::exception("NoiseRemoval") << "Error: Wrong channel groups for coherent noise removal, please check!";
      }

      for (int j=0; j<nSamples; j++) {
        holder[j] = holder[j] - cohNoise[chgroup][j];
      }

      if (channel==fCheckChannel) {
        for (int j=0; j<nSamples; j++) {
          h1_adc_raw_tailremoval_coherentremoval->SetBinContent(j+1, holder[j]);
        }
      }
    } // fRemoveCoherentNoise

    // deconvolute
    if (fDoDeconvolution) {
      holder.resize(transformSize); // 2048->4096, new elements are initialized as 0 here.
      sss->Deconvolute(channel, holder);
      holder.resize(fNticks); // 4096->2048, throw away the non-physical section
      SubtractBaseline(holder, fBaseSampleBins[plane]);
    } // fDoDeconvolution

    // apply wire-by-wire calibration
    if (fDodQdxCalib) {
      if (fdQdxCalib.find(channel) != fdQdxCalib.end()) {
        float constant = fdQdxCalib[channel];
        for (int j=0; j<nSamples; j++) {
          holder[j] *= constant;
        }
      }
    } // fDodQdxCalib

    // make a single ROI that spans the entire data size
    wirecol->push_back(recob::WireCreator(holder,*rawdigit).move());
    if (!util::CreateAssn(*this, e, *wirecol, rawdigit, *WireDigitAssn)) {
      throw art::Exception(art::errors::ProductRegistrationFailure)
      << "Can't associate wire #" << (wirecol->size() - 1)
      << " with raw digit #" << rawdigit.key();
    } // if failed to add association

  } // loop over rawdigitlist.size() i

  if (wirecol->size() == 0) {
    mf::LogWarning("CalWireArgoneut") << "No wires made for this event.";
  }

  e.put(std::move(wirecol));
  e.put(std::move(WireDigitAssn));

  return;
}


void NoiseRemoval::beginJob()
{
  mf::LogInfo("NoiseRemoval") << "beginJob";
  art::ServiceHandle<art::TFileService> tfs;
  
  
  h1_adc_raw = tfs->make<TH1D>("h1_adc_raw", Form("waveform on Channel %d; time [ticks]; ADC count", fCheckChannel), 2048, 0., 2048.);
  h1_adc_raw_tailremoval = tfs->make<TH1D>("h1_adc_raw_tailremoval", Form("waveform on Channel %d; time [ticks]; ADC count", fCheckChannel), 2048, 0., 2048.);
  h1_adc_raw_tailremoval_coherentremoval = tfs->make<TH1D>("h1_adc_raw_tailremoval_coherentremoval", Form("waveform on Channel %d; time [ticks]; ADC count", fCheckChannel), 2048, 0., 2048.);

  TFile file_response(fResponseFile.c_str());
  h1_decayconstant = (TH1D*)file_response.Get("real/decayHist");
}


std::vector<float> NoiseRemoval::getMedianCorrection(std::vector< std::vector<float> >& channelgroup, unsigned int dataSize, unsigned short plane)
{ 
  std::vector<float> correction(dataSize);
  std::vector< std::vector<float> > samples(dataSize); // [2048][channels in group with noise]

  // ignore the coherent noise if number of channels in a group less than 4
  if (channelgroup.size() <= 4) {
    for (size_t s = 0; s < samples.size(); ++s) {
      correction[s] = 0;
    }
    return correction;
  }
  
  for(size_t ch = 0; ch < channelgroup.size(); ++ch) {
    std::vector<float> thischannel = channelgroup[ch];
    auto mask = roiMask(thischannel, plane);

    for(size_t s = 0; s < samples.size(); ++s) {
      if (!mask[s]) { continue; }
      samples[s].push_back(thischannel[s]);
    }
  }

  for(size_t s = 0; s < samples.size(); ++s) {
    size_t n = samples[s].size();
    if(n < 2) { correction[s] = 0; continue; }
    std::sort(samples[s].begin(), samples[s].end());
    if((n % 2) == 0) { correction[s] = 0.5 * (samples[s][n/2] + samples[s][(n/2)-1]); }
    else { correction[s] = samples[s][(n-1)/2]; }
  }

  return correction;
} // getMedianCorrection


std::vector<bool> NoiseRemoval::roiMask(std::vector<float>& channel, unsigned short plane)
{
  // mask: true -> noise-like; false -> signal-like
  std::vector<bool> mask(channel.size(), true); // [2048]

  bool inroi = false;
  for (int i = 0; i < (int)channel.size(); ++i) {
    
    auto sig = channel[i]; // ADC at tick i

    if (inroi){
      if (sig > fRoiThresholdHigh[plane] || sig < -1.0*fRoiThresholdHigh[plane]) { 
        mask[i] = false; // signal-like
      }
      else {
        for (int p = 0; p <= fRoiPadHigh; ++p) {
          if ((i + p) < (int)mask.size())  mask[i + p] = false;
        }
        inroi = false;
      }   
    } // if (inroi)
    else {
      if (sig > fRoiThresholdLow[plane] || sig < -1.0*fRoiThresholdLow[plane]) {
        for (int p = fRoiPadLow; p >= 0; --p) {
          if (i - p >= 0) mask[i - p] = false;
        }
        inroi = true;
      }   
    }
  } // loop over channel.size() i

  return mask;
} // roiMask


void NoiseRemoval::estimatepars(std::vector<float> & holder, double &pedoffset, double &tail0, int channel)
{
  int nSamples = holder.size();
  // find t0 and p
  // tail: cofficient, let t0 be 1 and set all others to 0
  double temp_t0=1.;
  double temp_p = 0.;
  std::vector<float> temp_dat(2048, 0.); // d_i
  std::vector<float> temp_sig(2048); // s_i
  std::vector<float> temp_tail(2048); // t_i

  setData(temp_dat, temp_sig, temp_tail, temp_t0, temp_p, channel);

  std::vector<double> c_t(2048); // cofficients for t0
  for (int j=0; j<2048; j++) {
    c_t[j] = temp_sig[j]; // s_i
  }

  // ped: cofficient, let p be 1 and set all others to 0
  temp_t0 = 0.;
  temp_p = 1.;
  temp_dat.clear();
  temp_dat.resize(2048, 0.);
  temp_sig.clear();
  temp_sig.resize(2048, 0.);
  temp_tail.clear();
  temp_tail.resize(2048, 0.);

  setData(temp_dat, temp_sig, temp_tail, temp_t0, temp_p, channel);

  std::vector<double> c_p(2048);
  for (int j=0; j<2048; j++) {
    c_p[j] = temp_sig[j]; 
  }

  // data: cofficient, use data and set all others to be 0
  temp_t0 = 0.;
  temp_p = 0.;
  temp_dat.clear();
  temp_dat.resize(2048, 0.);
  //temp_sig.clear();
  //temp_sig.resize(2048, 0.);
  temp_tail.clear();
  temp_tail.resize(2048, 0.);

  setData(holder, temp_sig, temp_tail, temp_t0, temp_p, channel);

  std::vector<double> c_d(2048);

  for (int j=0; j<2048; j++) {
    c_d[j] = temp_sig[j]; 
  }

  // evaluate the K-params
  double kdd = 0.0;
  double kdt = 0.0;
  double kdp = 0.0;
  double ktt = 0.0;
  double ktp = 0.0;
  double kpp = 0.0;

  int nSamplesKeep = 0;

  for (int j=0; j<nSamples; j++) {
    if (std::abs(holder[j]) > fADCCut) continue;
    
    double cd = c_d[j];
    double ct = c_t[j];
    double cp = c_p[j];

    kdd += cd*cd;
    kdt += cd*ct;
    kdp += cd*cp;
    ktt += ct*ct;
    ktp += ct*cp;
    kpp += cp*cp;

    nSamplesKeep++;
  }

  double den =  ktt*kpp - ktp*ktp;

  double thisped, thistail0;
  if (den == 0) {
    cout << "\nWarning: denominartor is 0\n" << endl;
    thisped = 0.;
    thistail0 = 0.;
  }
  else {
    thistail0 = (kdp*ktp-kdt*kpp)/den;
    thisped = (kdt*ktp-kdp*ktt)/den;
  }
  
  pedoffset = thisped;
  tail0 = thistail0;

  if (channel == fCheckChannel) {
    double chsq = kdd + ktt*thistail0*thistail0 + kpp*thisped*thisped + 2.0*(kdt*thistail0 + kdp*thisped + ktp*thistail0*thisped);

    cout << "nSamplesKeep: " << nSamplesKeep << endl;
    cout << "chsq: " << chsq << endl;
    cout << "noise: " << sqrt(chsq/(nSamplesKeep-2)) << endl;
  }

} // estimatepars


void NoiseRemoval::setData(std::vector<float> & holder, std::vector<float> & signal, std::vector<float> & tail, double tail0, double ped, int channel)
{
  int nSamples = holder.size();

  double csum = tail0;
  double pedestal = ped;

  double beta = TMath::Exp(h1_decayconstant->GetBinContent(channel+1));
  double alpha = 1./beta - 1.;

  for (int j=0; j<nSamples; j++) {
    tail[j] = csum;
    signal[j] = holder[j] - tail[j] - pedestal; // noise is ignored
    csum = beta*tail[j] - alpha*beta*signal[j];
  }

} // setData


//////////////////////////////////////////////////////
void NoiseRemoval::SubtractBaseline(std::vector<float>& holder, int baseSampleBins)
{
  // subtract baseline using linear interpolation between regions defined
  // by the dataSize and baseSampleBins
  if(baseSampleBins <= 0) return;
  constexpr double fBaseVarCut = 20;
  // number of points to characterize the baseline
  unsigned short nBasePts = holder.size() / baseSampleBins;
  // the baseline offset vector
  std::vector<double> base(nBasePts, 0.);
  // find the average value in each region, using values that are
  // similar
  double fbins = baseSampleBins;
  unsigned short nfilld = 0;
  for(unsigned short ii = 0; ii < nBasePts; ++ii) {
    unsigned short loBin = ii * baseSampleBins;
    unsigned short hiBin = loBin + baseSampleBins;
    double ave = 0.;
    double sum = 0.;
    for(unsigned short bin = loBin; bin < hiBin; ++bin) {
      ave += holder[bin];
      sum += holder[bin] * holder[bin];
    } // bin
    ave = ave / fbins;
    double var = (sum - fbins * ave * ave) / (fbins - 1.);
    // Set the baseline for this region if the variance is small
    if(var < fBaseVarCut) {
      base[ii] = ave;
      ++nfilld;
    }
  } // ii

  // interpolate and subtract
  // find the first non-zero sample
  unsigned short thisRegion = 0;
  unsigned short nextRegion = 1;
  for(nextRegion = 1; nextRegion < base.size(); ++nextRegion) if(base[nextRegion] != 0) break;
  if(nextRegion == base.size()) return;
  double nBins = (nextRegion - thisRegion) * baseSampleBins;
  double slp = (base[nextRegion] - base[thisRegion]) / nBins;
  for(unsigned short bin = 0; bin < holder.size(); ++bin) {
    // determine which region we are in
    short region = bin / baseSampleBins;
    if(region == nextRegion) {
      // update the slope for the next region
      unsigned short nnr = 0;
      for(nnr = nextRegion + 1; nnr < base.size(); ++nnr) if(base[nnr] != 0) break;
      if(nnr == base.size()) nnr = base.size() - 1;
      thisRegion = nextRegion;
      nextRegion = nnr;
      if(thisRegion == nextRegion) return;
      nBins = (nextRegion - thisRegion) * baseSampleBins;
      slp = (base[nextRegion] - base[thisRegion]) / nBins;
    }
    double bs = base[thisRegion] + (double)(bin - thisRegion * baseSampleBins) * slp;

    holder[bin] -= bs;
  } // bin
} // SubtractBaseline
  


DEFINE_ART_MODULE(NoiseRemoval)
